#include <pic.h>				// pic specific identifiers
#define _XTAL_FREQ  4000000		// Xtal speed
__CONFIG(0x3D18);				// Config bits

#define LCD_PORT PORTA			// constants
#define LCD_TRIS TRISA			//
#define LCD_RS 	RA4
#define LCD_RW	RA6
#define LCD_E	RA7

#define KEY_PORT PORTB			// constants
#define KEY_TRIS TRISB			//
#define COL1 	RB0
#define COL2	RB1
#define COL3	RB2
#define COL4	RB3

#define CODE	2580			// set up a code

		// Required prototypes.. each function needs to be declared
		// if called BEFORE definition.

void LCD_init(void), LCD_cmd(unsigned char ch);
void LCD_goto(char line, char column), LCD_clr(void);
void LCD_cur(unsigned char ch), pulse_E(void), LCD_hex(int value);
void LCD_char(unsigned char ch), LCD_charD(unsigned char ch);
void LCD_printC(const char *str), LCD_printR(char *str);
void delayMs(int x);
unsigned char getkey(void);

unsigned char KEY_Table[] = {0x31, 0x32, 0x33, 0x46,
							0x34, 0x35, 0x36, 0x45,
							0x37, 0x38, 0x39, 0x44,
							0x41, 0x30, 0x42, 0x43};	// keypad numbers
unsigned char HEX_Table[] = {0x30, 0x31, 0x32, 0x33,
							0x34, 0x35, 0x36, 0x37,
							0x38, 0x39, 0x41, 0x42,
							0x43, 0x44, 0x45, 0x46}; 	//ASCII chracters

void main(void)						// program entry
	{
	int keycode = 0;
	unsigned char key = 0;			// new keypad variable
	char keycount = 0;
	CMCON = 0x7;					// Comparitors off
	LCD_TRIS = 0b00000000;			// Led port as outputs
	KEY_TRIS = 0x0F;				// column pins inputs
	delayMs(150);					// let LCD stabilise
	LCD_init();						// Initalise screen
	LCD_goto(1,0);					// line 1.
	LCD_printC("Enter Code!");
	LCD_goto(2,0);
	LCD_cur(1);						// set a flashing cursor
	while(1)						// endless Loop
		{		
		key = getkey();							// get a key
		if(key > 0 && key != 14)				// If its anything other than 0
			{
			LCD_char(KEY_Table[key-1]);
			keycount++;
			keycode *= 10;
			keycode += (KEY_Table[key-1] - 0x30) ;
			}
		if(key == 14)							// If its a 0
			{
			LCD_char(KEY_Table[key-1]);
			keycount++;
			keycode *= 10;
			}
		if(keycount == 4)						// four keys have been pressed.
			{
			LCD_goto(2,0);
			LCD_cur(0);
			if(CODE == keycode)					// Was it correct
				LCD_printC("Correct code!   ");
			else
				LCD_printC("Incorrect code! ");
			keycode = keycount = 0;
			LCD_goto(2,0);
			delayMs(1000);						// Wait a mo! for visual
			LCD_printC("                 ");
			LCD_goto(2,0);
			LCD_cur(1);
			}
		}
	}

unsigned char getkey()						//
	{
	char mask = 0xEF;						// mask for row selection
	char index;
	unsigned char key = 0;
	for(index = 0;index<4;index++)			//	Four rows
		{
		KEY_PORT = mask;
		if(!COL1) key=1;					// Check
		if(!COL2) key=2;					// each
		if(!COL3) key=3;					// column
		if(!COL4) key=4;
		if(key)
			{
			delayMs(300);
			return (index*4) + key;			// if a key was preesed
											// mulply rows by 4 and add
			}								// add the column.
		mask<<=1;							// next row.
		}
	return 0;
	}

void delayMs(int x)					// As __delay_ms() is small this is another
	{								// function
	while(x--)
		__delay_ms(1);
	}

void LCD_printC(const char * str)	// This passes the start a ROM character array
	{								// by default the pointer points to data section
	while(*str != 0)				// while the character pointed to isn't 0
		LCD_char(*str++);			// print out the character, then increment
	}								// the pointer down the array

void LCD_printR(char * str)	 		// This passes the start of a RAM character array
	{								// by default the pointer points to data section
	while(*str != 0)				// while the character pointed to isn't 0
		LCD_char(*str++);			// print out the character, then increment
	}								// the pointer down the array

void LCD_init()
	{
	LCD_cmd(0x20);					// 4 bit
	LCD_cmd(0x28);					// display shift
	LCD_cmd(0x6);					// character mode
	LCD_cmd(0xc);					// display on / off and cursor
	LCD_clr();						// clear display
	}


void LCD_hex(int value)
	{
	char data;								// dispay hex values.
	data = value >> 4 & 0xf;				// send upper nibble
	data = HEX_Table[data]; 				// hex lookup
	LCD_char(data);
	data = value & 0xf;						// send lower nibble
	data = HEX_Table[data]; 				// hex lookup
	LCD_char(data);
	}

void LCD_goto(char line, char column)		// combines line and lineW
	{
	unsigned char data = 0x80;				// default to 1
	if(line == 2)data = 0xc0;				// change to 2
	data += column;							// add in  column
	LCD_cmd(data);
	}

void LCD_clr()
	{
	LCD_cmd(1);								// Clr screen
	}

void LCD_cur(char on)
	{
	unsigned char cur = 0xc;				// cursor off
	if(on) cur = 0xd;						// cursor on
	LCD_cmd(cur);
	}

void LCD_cmd(unsigned char ch)
	{
	LCD_PORT = ch >> 4 & 0xf;			// write high nibble
	LCD_RS = 0;
	pulse_E();
	LCD_PORT = ch & 0xf;				// write low nibble
	LCD_RS = 0;
	pulse_E();
	__delay_ms(5);
	}

void LCD_charD(unsigned char ch)
	{
	ch+=0x30;
	LCD_char(ch);						// convert to ascii
	}

void LCD_char(unsigned char ch)
	{
	LCD_PORT = ch >> 4 & 0xf;			// High nibble
	LCD_RS = 1;
	pulse_E();
	LCD_PORT = ch & 0xf;				// low nibble
	LCD_RS = 1;
	pulse_E();
	__delay_ms(5);
	}

void pulse_E()
	{
	LCD_E = 1;
	__delay_us(1);
	LCD_E = 0;
	}