#include <pic.h>				// pic specific identifiers
#define _XTAL_FREQ  4000000		// Xtal speed
__CONFIG(0x3D18);				// Config bits

#define RS232_PORT PORTA			// constants
#define RS232_TRIS TRISA			//
#define TX 	RA6
#define RX	RA7


#define KEY_PORT PORTB			// constants
#define KEY_TRIS TRISB			//
#define COL1 	RB0
#define COL2	RB1
#define COL3	RB2
#define COL4	RB3
#define BAUD		84				// 1/9600 = 104us ( 20us for loop and math)
#define START_BIT	BAUD / 2		// half bit

#define CODE	2580			// set up a code

		// Required prototypes.. each function needs to be declared
		// if called BEFORE definition.

void Serinit(void), SerStringC(const char* str);
void Serout( unsigned char ch), newLine(void);
unsigned char Serin(void);
void Ser_hex(int value), delayMs(int x);
unsigned char getkey(void);

unsigned char KEY_Table[] = {0x31, 0x32, 0x33, 0x46,
							0x34, 0x35, 0x36, 0x45,
							0x37, 0x38, 0x39, 0x44,
							0x41, 0x30, 0x42, 0x43};	// keypad numbers
unsigned char HEX_Table[] = {0x30, 0x31, 0x32, 0x33,
							0x34, 0x35, 0x36, 0x37,
							0x38, 0x39, 0x41, 0x42,
							0x43, 0x44, 0x45, 0x46}; 	//ASCII chracters

void main(void)						// program entry
	{
	int keycode = 0;
	unsigned char key = 0;			// new keypad variable
	char keycount = 0;
	CMCON = 0x7;					// Comparitors off
	KEY_TRIS = 0x0F;				// column pins inputs
	delayMs(150);					// let LCD stabilise
	Serinit();
	SerStringC("Keypad test!.");
	newLine();
	while(1)							// endless Loop
		{
		key = getkey();
		if(key)							// A key was pressed
			{
			SerStringC("Hex ");			// First the hexidecimal
			Ser_hex(key);	
			SerStringC(" ASCII ");		// Then the ASCII equivelent,
			Serout(KEY_Table[key-1]); 	// We dont have a zero.
			newLine();					// next line
			}
		}						
	}
void newLine()
	{
	Serout(0xa);
	Serout(0xd);
	}

void SerStringC(const char* str)
	{
	while(*str != 0)
		Serout(*str++);
	}

void Ser_hex(int value)
	{
	char data;								// dispay hex values.
	data = value >> 4 & 0xf;				// send upper nibble
	data = HEX_Table[data]; 				// hex lookup
	Serout(data);
	data = value & 0xf;						// send lower nibble
	data = HEX_Table[data]; 				// hex lookup
	Serout(data);
	}
void Serinit()
	{
	RS232_TRIS = 0b10000000;			// RS232  port
	TX = 1;
	}

unsigned char Serin()					// wait and get a byte on the serial
	{
	char index, ch=0;
	while(1)
		{
		while(RX);						// wait for start bit
		__delay_us(START_BIT);			// Half bit delay
		if(!RX)							// still start?
			{
			__delay_us(BAUD);				// bit delay
			for(index = 0;index<7;index++)
				{
				if(RX)						// bit or no bit
					ch+=0x80;				// add MSB
				__delay_us(BAUD);			// bit delay
				ch>>=1;						// shift along
				}							// until done
			if(RX) ch+=0x80;				// last bit
			return ch;						// return character
			}								// not bothered about the stop bit
		}
	return 0;
	}

void Serout( unsigned char ch)
	{
	char index;
	TX = 0;									// send a start bit
	__delay_us(BAUD);
	for(index = 0; index<8 ; index++)		// send the rest
		{
		if(ch & 0x1)						// test bit
			TX = 1;							// send high
		else
			TX = 0;							// send low
		__delay_us(BAUD);					// bit delay
		ch >>= 1;
		}
	TX = 1;									// send a stop bit
	__delay_us(BAUD);
	}
unsigned char getkey()						//
	{
	char mask = 0xEF;						// mask for row selection
	char index;
	unsigned char key = 0;
	for(index = 0;index<4;index++)			//	Four rows
		{
		KEY_PORT = mask;
		if(!COL1) key=1;					// Check
		if(!COL2) key=2;					// each
		if(!COL3) key=3;					// column
		if(!COL4) key=4;
		if(key)
			{
			delayMs(300);
			return (index*4) + key;			// if a key was preesed
											// mulply rows by 4 and add
			}								// add the column.
		mask<<=1;							// next row.
		}
	return 0;
	}

void delayMs(int x)					// As __delay_ms() is small this is another
	{								// function
	while(x--)
		__delay_ms(1);
	}

